"""Exposes common utilities.

"""
import typing
from typing import Any, Tuple, Optional, Union, Dict, Type, List

import orjson


def strip_leading(word: str, substring: str) -> str:
    """Strips the leading substring if it exists.

    This is contrary to rstrip which removes each character in the substring

    Args:
        word: the string to strip from
        substring: the string to be stripped from the word.

    Returns:
        the stripped word
    """
    if word.startswith(substring):
        return word[len(substring) :]
    return word


def typing_get_args(v: Any) -> Tuple[Any, ...]:
    """Gets the __args__ of the annotations of a given typing.

    Args:
        v: the typing object whose __args__ are required.

    Returns:
        the __args__ of the item passed
    """
    try:
        return typing.get_args(v)
    except AttributeError:
        return getattr(v, "__args__", ()) if v is not typing.Generic else typing.Generic


def typing_get_origin(v: Any) -> Optional[Any]:
    """Gets the __origin__ of the annotations of a given typing.

    Args:
        v: the typing object whose __origin__ are required.

    Returns:
        the __origin__ of the item passed
    """
    try:
        return typing.get_origin(v)
    except AttributeError:
        return getattr(v, "__origin__", None)


def from_bytes_to_str(value: Union[str, bytes]) -> str:
    """Converts bytes to str.

    Args:
        value: the potentially bytes object to transform.

    Returns:
        the string value of the argument passed
    """
    try:
        return str(value, "utf-8")
    except TypeError:
        return value


def from_str_or_bytes_to_any(value: Any, field_type: Type) -> Any:
    """Converts str or bytes to arbitrary data.

    Converts the `value` from a string or bytes to the `field_type`.

    Args:
        value: the string or bytes to be transformed to the `field_type`
        field_type: the type to which value is to be converted

    Returns:
        the `field_type` version of the `value`.
    """
    if isinstance(value, (bytes, bytearray, memoryview)):
        # bytearray-like objects should always be JSON-parsed
        return orjson.loads(value)

    elif field_type == Optional[str] and value != "null":
        # optional strings have 'null' when None
        return value

    elif field_type == str:
        return value

    elif not isinstance(value, str):
        return value

    # JSON parse all other values that are str
    return orjson.loads(value)


def from_any_to_valid_redis_type(value: Any) -> Union[str, bytes, List[Any]]:
    """Converts arbitrary data into valid redis types

    Converts the the `value` from any type to a type that
    are acceptable by redis.
    By default, complex data is transformed to bytes.

    Args:
        value: the value to be transformed to a valid redis data type

    Returns:
        the transformed version of the `value`.
    """
    if isinstance(value, str):
        return value
    elif isinstance(value, set):
        return list(value)
    return orjson.dumps(value, default=default_json_dump)


def default_json_dump(obj: Any):
    """Serializes objects orjson cannot serialize.

    Args:
        obj: the object to serialize

    Returns:
        the bytes or string value of the object
    """
    try:
        return obj.model_dump_json()
    except AttributeError:
        return obj


def from_dict_to_key_value_list(data: Dict[str, Any]) -> List[Any]:
    """Converts dict to flattened list of key, values.

    {"foo": "bar", "hen": "rooster"} becomes ["foo", "bar", "hen", "rooster"]
    When redis lua scripts are used, the value returned is a flattened list,
    similar to that shown above.

    Args:
        data: the dictionary to flatten

    Returns:
        the flattened list version of `data`
    """
    parsed_list = []

    for k, v in data.items():
        parsed_list.append(k)
        parsed_list.append(v)

    return parsed_list



"""Asynchronous API for pydantic-redis ORM.

Typical usage example:

```python
import asyncio
from pydantic_redis.asyncio import Store, Model, RedisConfig

class Book(Model):
    _primary_key_field = 'title'
    title: str

async def main():
    store = Store(name="sample", redis_config=RedisConfig())
    store.register_model(Book)

    await Book.insert(Book(title="Oliver Twist", author="Charles Dickens"))
    await Book.update(
        _id="Oliver Twist", data={"author": "Jane Austen"}, life_span_seconds=3600
    )
    results = await Book.select()
    await Book.delete(ids=["Oliver Twist", "Great Expectations"])

if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())
```
"""

from .model import Model
from .store import Store
from ..config import RedisConfig

__all__ = [Model, Store, RedisConfig]



"""Exposes utils for getting properties of the Model

Attributes:
    NESTED_MODEL_PREFIX (str): the prefix for fields with single nested models
    NESTED_MODEL_LIST_FIELD_PREFIX (str): the prefix for fields with lists of nested models
    NESTED_MODEL_TUPLE_FIELD_PREFIX (str): the prefix for fields with tuples of nested models
"""

from typing import Type, Any

from .base import AbstractModel

NESTED_MODEL_PREFIX = "__"
NESTED_MODEL_LIST_FIELD_PREFIX = "___"
NESTED_MODEL_TUPLE_FIELD_PREFIX = "____"


def get_redis_key(model: Type[AbstractModel], primary_key_value: Any):
    """Gets the key used internally in redis for the `primary_key_value` of `model`.

    Args:
        model: the model for which the key is to be generated
        primary_key_value: the external facing primary key value

    Returns:
        the primary key internally used for `primary_key_value` of `model`
    """
    return f"{get_redis_key_prefix(model)}{primary_key_value}"


def get_redis_key_prefix(model: Type[AbstractModel]):
    """Gets the prefix for keys used internally in redis for records of `model`.

    Args:
        model: the model for which the redis key prefix is to be generated

    Returns:
        the prefix of the all the redis keys that are associated with this model
    """
    model_name = model.__name__.lower()
    return f"{model_name}_%&_"


def get_redis_keys_regex(model: Type[AbstractModel]):
    """Gets the regex for all keys of records of `model` used internally in redis.

    Args:
        model: the model for which the redis key regex is to be generated

    Returns:
        the regular expression for all keys of records of `model` used internally in redis
    """
    return f"{get_redis_key_prefix(model)}*"


def get_model_index_key(model: Type[AbstractModel]):
    """Gets the key for the index set of `model` used internally in redis.

    The index for each given model stores the primary keys for all records
    that belong to the given model.

    Args:
        model: the model whose index is wanted.

    Returns:
        the the key for the index set of `model` used internally in redis.
    """
    table_name = model.__name__.lower()
    return f"{table_name}__index"



"""Exposes the Base `Model` class for creating custom synchronous models.

This module contains the `Model` class which should be inherited when
creating model's for use in the synchronous API of pydantic-redis
"""
from typing import Optional, List, Any, Union, Dict

from .._shared.model import AbstractModel
from .._shared.model.insert_utils import insert_on_pipeline
from .._shared.model.select_utils import (
    select_all_fields_all_ids,
    select_all_fields_some_ids,
    select_some_fields_all_ids,
    select_some_fields_some_ids,
    parse_select_response,
)

from .store import Store
from .._shared.model.delete_utils import delete_on_pipeline


class Model(AbstractModel):
    """
    The Base class for all Synchronous models.

    Inherit this class when creating a new model.
    The new model should have `_primary_key_field` defined.
    Any interaction with redis is done through `Model`'s.
    """

    _store: Store

    @classmethod
    def insert(
        cls,
        data: Union[List[AbstractModel], AbstractModel],
        life_span_seconds: Optional[float] = None,
    ):
        """Inserts a given record or list of records into the redis.

        Can add a single record or multiple records into redis.
        The records must be instances of this class. i.e. a `Book`
        model can only insert `Book` instances.

        Args:
            data: a model instance or list of model instances to put
                into the redis store
            life_span_seconds: the time-to-live in seconds of the records
                to be inserted. If not specified, it defaults to the `Store`'s
                life_span_seconds.
        """
        store = cls.get_store()

        life_span = (
            life_span_seconds
            if life_span_seconds is not None
            else store.life_span_in_seconds
        )
        with store.redis_store.pipeline(transaction=True) as pipeline:
            data_list = []

            if isinstance(data, list):
                data_list = data
            elif isinstance(data, AbstractModel):
                data_list = [data]

            for record in data_list:
                insert_on_pipeline(
                    model=cls,
                    _id=None,
                    pipeline=pipeline,
                    record=record,
                    life_span=life_span,
                )

            return pipeline.execute()

    @classmethod
    def update(
        cls, _id: Any, data: Dict[str, Any], life_span_seconds: Optional[float] = None
    ):
        """Updates the record whose primary key is `_id`.

        Updates the record of this Model in redis whose primary key is equal to the `_id` provided.
        The record is partially updated from the `data`.
        If `life_span_seconds` is provided, it will also update the time-to-live of
        the record.

        Args:
            _id: the primary key of record to be updated.
            data: the new changes
            life_span_seconds: the new time-to-live for the record
        """
        store = cls.get_store()
        life_span = (
            life_span_seconds
            if life_span_seconds is not None
            else store.life_span_in_seconds
        )
        with store.redis_store.pipeline(transaction=True) as pipeline:
            if isinstance(data, dict):
                insert_on_pipeline(
                    model=cls,
                    _id=_id,
                    pipeline=pipeline,
                    record=data,
                    life_span=life_span,
                )

            return pipeline.execute()

    @classmethod
    def delete(cls, ids: Union[Any, List[Any]]):
        """Removes a list of this Model's records from redis

        Removes all the records for the current Model whose primary keys
        have been included in the `ids` passed.

        Args:
            ids: list of primary keys of the records to remove
        """
        store = cls.get_store()
        with store.redis_store.pipeline() as pipeline:
            delete_on_pipeline(model=cls, pipeline=pipeline, ids=ids)
            return pipeline.execute()

    @classmethod
    def select(
        cls,
        columns: Optional[List[str]] = None,
        ids: Optional[List[Any]] = None,
        skip: int = 0,
        limit: Optional[int] = None,
        **kwargs,
    ) -> Union["Model", Dict[str, Any]]:
        """Retrieves records of this Model from redis.

        Retrieves the records for this Model from redis.

        Args:
            columns: the fields to return for each record
            ids: the primary keys of the records to returns
            skip: the number of records to skip. (default: 0)
            limit: the maximum number of records to return

        Returns:
            By default, it returns all records that belong to current Model.

            If `ids` are specified, it returns only records whose primary keys
            have been listed in `ids`.

            If `skip` and `limit` are specified WITHOUT `ids`, a slice of
            all records are returned.

            If `limit` and `ids` are specified, `limit` is ignored.

            If `columns` are specified, a list of dictionaries containing only
            the fields specified in `columns` is returned. Otherwise, instances
            of the current Model are returned.
        """
        if columns is None and ids is None:
            response = select_all_fields_all_ids(model=cls, skip=skip, limit=limit)

        elif columns is None and isinstance(ids, list):
            response = select_all_fields_some_ids(model=cls, ids=ids)

        elif isinstance(columns, list) and ids is None:
            response = select_some_fields_all_ids(
                model=cls, fields=columns, skip=skip, limit=limit
            )

        elif isinstance(columns, list) and isinstance(ids, list):
            response = select_some_fields_some_ids(model=cls, fields=columns, ids=ids)

        else:
            raise ValueError(
                f"columns {columns}, ids: {ids} should be either None or lists"
            )

        return parse_select_response(
            model=cls, response=response, as_models=(columns is None)
        )


Store.model_rebuild()



"""Synchronous API for pydantic-redis ORM.

Typical usage example:

```python
# from pydantic_redis import Store, Model, RedisConfig
from pydantic_redis.syncio import Store, Model, RedisConfig

class Book(Model):
    _primary_key_field = 'title'
    title: str

if __name__ == '__main__':
    store = Store(name="sample", redis_config=RedisConfig())
    store.register_model(Book)

    Book.insert(Book(title="Oliver Twist", author="Charles Dickens"))
    Book.update(
        _id="Oliver Twist", data={"author": "Jane Austen"}, life_span_seconds=3600
    )
    results = Book.select()
    Book.delete(ids=["Oliver Twist", "Great Expectations"])
```
"""

from .model import Model
from .store import Store
from ..config import RedisConfig

__all__ = [Model, Store, RedisConfig]



"""Exposes the `Store` for managing a collection of asynchronous Model's.

Stores represent a collection of different kinds of records saved in
a redis database. They only expose records whose `Model`'s have been
registered in them. Thus redis server can have multiple stores each
have a different image of the actual data in redis.

A model must be registered with a store before it can interact with
a redis database.
"""
from typing import Dict, Type, TYPE_CHECKING

from redis import asyncio as redis

from .._shared.store import AbstractStore

if TYPE_CHECKING:
    from .model import Model


class Store(AbstractStore):
    """Manages a collection of Model's, connecting them to a redis database

    A Model can only interact with a redis database when it is registered
    with a `Store` that is connected to that database.

    Attributes:
        models (Dict[str, Type[pydantic_redis.syncio.Model]]): a mapping of registered `Model`'s, with the keys being the
            Model name
        name (str): the name of this Store
        redis_config (pydantic_redis.syncio.RedisConfig): the configuration for connecting to a redis database
        redis_store (Optional[redis.Redis]): an Redis instance associated with this store (default: None)
        life_span_in_seconds (Optional[int]): the default time-to-live for the records inserted in this store
            (default: None)
    """

    models: Dict[str, Type["Model"]] = {}

    def _connect_to_redis(self) -> redis.Redis:
        """Connects the store to redis.

        See the base class.
        """
        return redis.from_url(
            self.redis_config.redis_url,
            encoding=self.redis_config.encoding,
            decode_responses=True,
        )



"""Exposes utilities for selecting records from redis using lua scripts.

"""
from typing import List, Any, Type, Union, Awaitable, Optional

from pydantic_redis._shared.model.prop_utils import (
    NESTED_MODEL_PREFIX,
    NESTED_MODEL_LIST_FIELD_PREFIX,
    NESTED_MODEL_TUPLE_FIELD_PREFIX,
    get_redis_keys_regex,
    get_redis_key_prefix,
    get_model_index_key,
)


from .base import AbstractModel


def get_select_fields(model: Type[AbstractModel], columns: List[str]) -> List[str]:
    """Gets the fields to be used for selecting HMAP fields in Redis.

    It replaces any fields in `columns` that correspond to nested records with their
    `__` prefixed versions.

    Args:
        model: the model for which the fields for selecting are to be derived.
        columns: the fields that are to be transformed into fields for selecting.

    Returns:
        the fields for selecting, with nested fields being given appropriate prefixes.
    """
    fields = []
    nested_model_list_fields = model.get_nested_model_list_fields()
    nested_model_tuple_fields = model.get_nested_model_tuple_fields()
    nested_model_fields = model.get_nested_model_fields()

    for col in columns:

        if col in nested_model_fields:
            col = f"{NESTED_MODEL_PREFIX}{col}"
        elif col in nested_model_list_fields:
            col = f"{NESTED_MODEL_LIST_FIELD_PREFIX}{col}"
        elif col in nested_model_tuple_fields:
            col = f"{NESTED_MODEL_TUPLE_FIELD_PREFIX}{col}"

        fields.append(col)
    return fields


def select_all_fields_all_ids(
    model: Type[AbstractModel],
    skip: int = 0,
    limit: Optional[int] = None,
) -> Union[List[List[Any]], Awaitable[List[List[Any]]]]:
    """Retrieves all records of the given model in the redis database.

    Args:
        model: the Model whose records are to be retrieved.
        skip: the number of records to skip.
        limit: the maximum number of records to return. If None, limit is infinity.

    Returns:
        the list of records from redis, each record being a flattened list of key-values.
        In case we are using async, an Awaitable of that list is returned instead.
    """
    if isinstance(limit, int):
        return _select_all_ids_all_fields_paginated(model=model, limit=limit, skip=skip)
    else:
        table_keys_regex = get_redis_keys_regex(model=model)
        args = [table_keys_regex]
        store = model.get_store()
        return store.select_all_fields_for_all_ids_script(args=args)


def select_all_fields_some_ids(
    model: Type[AbstractModel], ids: List[str]
) -> Union[List[List[Any]], Awaitable[List[List[Any]]]]:
    """Retrieves some records from redis.

    Args:
        model: the Model whose records are to be retrieved.
        ids: the list of primary keys of the records to be retrieved.

    Returns:
        the list of records where each record is a flattened key-value list.
        In case we are using async, an Awaitable of that list is returned instead.
    """
    table_prefix = get_redis_key_prefix(model=model)
    keys = [f"{table_prefix}{key}" for key in ids]
    store = model.get_store()
    return store.select_all_fields_for_some_ids_script(keys=keys)


def select_some_fields_all_ids(
    model: Type[AbstractModel],
    fields: List[str],
    skip: int = 0,
    limit: Optional[int] = None,
) -> Union[List[List[Any]], Awaitable[List[List[Any]]]]:
    """Retrieves records of model from redis, each as with a subset of the fields.

    Args:
        model: the Model whose records are to be retrieved.
        fields: the subset of fields to return for each record.
        skip: the number of records to skip.
        limit: the maximum number of records to return. If None, limit is infinity.

    Returns:
        the list of records from redis, each record being a flattened list of key-values.
        In case we are using async, an Awaitable of that list is returned instead.
    """
    columns = get_select_fields(model=model, columns=fields)

    if isinstance(limit, int):
        return _select_some_fields_all_ids_paginated(
            model=model, columns=columns, limit=limit, skip=skip
        )
    else:
        table_keys_regex = get_redis_keys_regex(model=model)
        args = [table_keys_regex, *columns]
        store = model.get_store()
        return store.select_some_fields_for_all_ids_script(args=args)


def select_some_fields_some_ids(
    model: Type[AbstractModel], fields: List[str], ids: List[str]
) -> Union[List[List[Any]], Awaitable[List[List[Any]]]]:
    """Retrieves some records of current model from redis, each as with a subset of the fields.

    Args:
        model: the Model whose records are to be retrieved.
        fields: the subset of fields to return for each record.
        ids: the list of primary keys of the records to be retrieved.

    Returns:
        the list of records from redis, each record being a flattened list of key-values.
        In case we are using async, an Awaitable of that list is returned instead.
    """
    table_prefix = get_redis_key_prefix(model=model)
    keys = [f"{table_prefix}{key}" for key in ids]
    columns = get_select_fields(model=model, columns=fields)
    store = model.get_store()
    return store.select_some_fields_for_some_ids_script(keys=keys, args=columns)


def parse_select_response(
    model: Type[AbstractModel], response: List[List], as_models: bool
):
    """Casts a list of flattened key-value lists into a list of models or dicts.

    It replaces any foreign keys with the related model instances,
    and converts the list of flattened key-value lists into a list of models or dicts.
    e.g. [["foo", "bar", "head", 9]] => [{"foo": "bar", "head": 9}]

    Args:
        model: the model to which the items in `response` belong
        response: the response from redis lua scripts that is to be cast.
        as_models: the flag that determines whether to cast items as models
            or as dicts.

    Returns:
        If `as_models` is true, list of models else list of dicts
    """
    if len(response) == 0:
        return None

    if as_models:
        return [
            model(**model.deserialize_partially(record))
            for record in response
            if record != []
        ]

    return [model.deserialize_partially(record) for record in response if record != []]


def _select_all_ids_all_fields_paginated(
    model: Type[AbstractModel], limit: int, skip: Optional[int]
):
    """Retrieves a slice of all records of the given model in the redis database.

    Args:
        model: the Model whose records are to be retrieved.
        skip: the number of records to skip.
        limit: the maximum number of records to return. If None, limit is infinity.

    Returns:
        the list of records from redis, each record being a flattened list of key-values.
        In case we are using async, an Awaitable of that list is returned instead.
    """
    if skip is None:
        skip = 0
    table_index_key = get_model_index_key(model)
    args = [table_index_key, skip, limit]
    store = model.get_store()
    return store.paginated_select_all_fields_for_all_ids_script(args=args)


def _select_some_fields_all_ids_paginated(
    model: Type[AbstractModel], columns: List[str], limit: int, skip: int
):
    """Retrieves a slice of all records of model from redis, each as with a subset of the fields.

    Args:
        model: the Model whose records are to be retrieved.
        columns: the subset of fields to return for each record.
        skip: the number of records to skip.
        limit: the maximum number of records to return. If None, limit is infinity.

    Returns:
        the list of records from redis, each record being a flattened list of key-values.
        In case we are using async, an Awaitable of that list is returned instead.
    """
    if skip is None:
        skip = 0
    table_index_key = get_model_index_key(model)
    args = [table_index_key, skip, limit, *columns]
    store = model.get_store()
    return store.paginated_select_some_fields_for_all_ids_script(args=args)



"""Exposes the configuration for connecting to a redis database.
"""
from typing import Optional

from pydantic import ConfigDict, BaseModel


class RedisConfig(BaseModel):
    """Configuration for connecting to redis database.

    Inorder to connect to a redis database, there are a number of
    configurations that are needed including the server's host address
    and port. `RedisConfig` computes a redis-url similar to
    `redis://:password@host:self.port/db`

    Attributes:
        host (str): the host address where the redis server is found (default: 'localhost').
        port (int): the port on which the redis server is running (default: 6379).
        db (int): the redis database identifier (default: 0).
        password (Optional[int]): the password for connecting to the
            redis server (default: None).
        ssl (bool): whether the connection to the redis server is to be via TLS (default: False)
        encoding: (Optional[str]): the string encoding used with the redis database
            (default: utf-8)
    """

    model_config = ConfigDict(from_attributes=True)

    host: str = "localhost"
    port: int = 6379
    db: int = 0
    password: Optional[str] = None
    ssl: bool = False
    encoding: Optional[str] = "utf-8"

    @property
    def redis_url(self) -> str:
        """a redis URL of form `redis://:password@host:port/db`. (`rediss://..` if TLS)."""
        proto = "rediss" if self.ssl else "redis"
        if self.password is None:
            return f"{proto}://{self.host}:{self.port}/{self.db}"
        return f"{proto}://:{self.password}@{self.host}:{self.port}/{self.db}"



"""Exposes the utility functions for inserting records into redis.

"""
from datetime import datetime
from typing import Union, Optional, Any, Dict, Tuple, List, Type

from redis.asyncio.client import Pipeline as AioPipeline
from redis.client import Pipeline

from .prop_utils import (
    get_redis_key,
    get_model_index_key,
    NESTED_MODEL_PREFIX,
    NESTED_MODEL_LIST_FIELD_PREFIX,
    NESTED_MODEL_TUPLE_FIELD_PREFIX,
)

from .base import AbstractModel


def insert_on_pipeline(
    model: Type[AbstractModel],
    pipeline: Union[Pipeline, AioPipeline],
    _id: Optional[Any],
    record: Union[AbstractModel, Dict[str, Any]],
    life_span: Optional[Union[float, int]] = None,
) -> Any:
    """Add an insert operation to the redis pipeline.

    Later when the pipeline.execute is called, the actual inserts occur.
    This reduces the number of network requests to the redis server thus
    improving performance.

    Args:
        model: the Model whose records are to be inserted into redis.
        pipeline: the Redis pipeline on which the insert operation
            is to be added.
        _id: the primary key of the record to be inserted in redis.
            It is None when inserting, and some value when updating.
        record: the model instance or dictionary to be inserted into redis.
        life_span: the time-to-live in seconds for the record to be inserted.
            (default: None)

    Returns:
        the primary key of the record that is to be inserted.
    """
    key = _id if _id is not None else getattr(record, model.get_primary_key_field())
    data = _serialize_nested_models(
        model=model, pipeline=pipeline, record=record, life_span=life_span
    )
    name = get_redis_key(model=model, primary_key_value=key)
    mapping = model.serialize_partially(data)
    pipeline.hset(name=name, mapping=mapping)

    if life_span is not None:
        pipeline.expire(name=name, time=life_span)
    # save the primary key in an index: a sorted set, whose score is current timestamp
    table_index_key = get_model_index_key(model)
    timestamp = datetime.utcnow().timestamp()
    pipeline.zadd(table_index_key, {name: timestamp})
    if life_span is not None:
        pipeline.expire(table_index_key, time=life_span)

    return name


def _serialize_nested_models(
    model: Type[AbstractModel],
    pipeline: Union[Pipeline, AioPipeline],
    record: Union[AbstractModel, Dict[str, Any]],
    life_span: Optional[Union[float, int]] = None,
) -> Dict[str, Any]:
    """Converts nested models into their primary keys.

    In order to make the record serializable, all nested models including those in
    lists and tuples of nested models are converted to their primary keys,
    after being their insert operations have been added to the pipeline.

    A few cleanups it does include:
      - Upserting any nested records in `record`
      - Replacing the keys of nested records with their NESTED_MODEL_PREFIX prefixed versions
        e.g. `__author` instead of author
      - Replacing the keys of lists of nested records with their NESTED_MODEL_LIST_FIELD_PREFIX prefixed versions
        e.g. `__%&l_author` instead of author
      - Replacing the keys of tuples of nested records with their NESTED_MODEL_TUPLE_FIELD_PREFIX prefixed versions
        e.g. `__%&l_author` instead of author
      - Replacing the values of nested records with their foreign keys

    Args:
        model: the model the given record belongs to.
        pipeline: the redis pipeline on which the redis operations are to be done.
        record: the model or dictionary whose nested models are to be serialized.
        life_span: the time-to-live in seconds for the given record (default: None).

    Returns:
        the partially serialized dict that has no nested models
    """
    data = record.items() if isinstance(record, dict) else record
    new_data = {}

    nested_model_list_fields = model.get_nested_model_list_fields()
    nested_model_tuple_fields = model.get_nested_model_tuple_fields()
    nested_model_fields = model.get_nested_model_fields()

    for k, v in data:
        key, value = k, v

        if key in nested_model_list_fields:
            key, value = _serialize_list(
                key=key, value=value, pipeline=pipeline, life_span=life_span
            )
        elif key in nested_model_tuple_fields:
            key, value = _serialize_tuple(
                key=key,
                value=value,
                pipeline=pipeline,
                life_span=life_span,
                tuple_fields=nested_model_tuple_fields,
            )
        elif key in nested_model_fields:
            key, value = _serialize_model(
                key=key, value=value, pipeline=pipeline, life_span=life_span
            )

        new_data[key] = value
    return new_data


def _serialize_tuple(
    key: str,
    value: Tuple[AbstractModel],
    pipeline: Union[Pipeline, AioPipeline],
    life_span: Optional[Union[float, int]],
    tuple_fields: Dict[str, Tuple[Any, ...]],
) -> Tuple[str, List[Any]]:
    """Replaces models in a tuple with strings.

    It adds insert operations for the records in the tuple onto the pipeline
    and returns the tuple with the models replaced by their primary keys as value.

    Returns:
        key: the original `key` prefixed with NESTED_MODEL_TUPLE_FIELD_PREFIX
        value: tthe tuple with the models replaced by their primary keys
    """
    try:
        field_types = tuple_fields.get(key, ())
        value = [
            insert_on_pipeline(
                model=field_type,
                _id=None,
                pipeline=pipeline,
                record=item,
                life_span=life_span,
            )
            if issubclass(field_type, AbstractModel)
            else item
            for field_type, item in zip(field_types, value)
        ]
        key = f"{NESTED_MODEL_TUPLE_FIELD_PREFIX}{key}"
    except TypeError:
        # In case the value is None, just ignore
        pass

    return key, value


def _serialize_list(
    key: str,
    value: List[AbstractModel],
    pipeline: Union[Pipeline, AioPipeline],
    life_span: Optional[Union[float, int]],
) -> Tuple[str, List[Any]]:
    """Casts a list of models into a list of strings

    It adds insert operations for the records in the list onto the pipeline
    and returns a list of their primary keys as value.

    Returns:
        key: the original `key` prefixed with NESTED_MODEL_LIST_FIELD_PREFIX
        value: the list of primary keys of the records to be inserted
    """
    try:
        value = [
            insert_on_pipeline(
                model=item.__class__,
                _id=None,
                pipeline=pipeline,
                record=item,
                life_span=life_span,
            )
            for item in value
        ]
        key = f"{NESTED_MODEL_LIST_FIELD_PREFIX}{key}"
    except TypeError:
        # In case the value is None, just ignore
        pass

    return key, value


def _serialize_model(
    key: str,
    value: AbstractModel,
    pipeline: Union[Pipeline, AioPipeline],
    life_span: Optional[Union[float, int]],
) -> Tuple[str, str]:
    """Casts a model into a string

    It adds an insert operation for the given model onto the pipeline
    and returns its primary key as value.

    Returns:
        key: the original `key` prefixed with NESTED_MODEL_PREFIX
        value: the primary key of the model
    """
    try:
        value = insert_on_pipeline(
            model=value.__class__,
            _id=None,
            pipeline=pipeline,
            record=value,
            life_span=life_span,
        )
        key = f"{NESTED_MODEL_PREFIX}{key}"
    except TypeError:
        # In case the value is None, just ignore
        pass

    return key, value



"""Exposes shared utilities for deleting records from redis

"""
from typing import Type, Union, List

from redis.client import Pipeline
from redis.asyncio.client import Pipeline as AioPipeline

from pydantic_redis._shared.model import AbstractModel
from pydantic_redis._shared.model.prop_utils import get_redis_key, get_model_index_key


def delete_on_pipeline(
    model: Type[AbstractModel], pipeline: Union[Pipeline, AioPipeline], ids: List[str]
):
    """Adds delete operations for the given ids to the redis pipeline.

    Args:
        model: the Model from which the given records are to be deleted.
        pipeline: the Redis pipeline on which the delete operations are
            to be added.
        ids: the list of primary keys of the records that are to be removed.

    Later when pipeline.execute is called later, the actual deletion occurs
    """
    primary_keys = []

    if isinstance(ids, list):
        primary_keys = ids
    elif ids is not None:
        primary_keys = [ids]

    names = [
        get_redis_key(model=model, primary_key_value=primary_key_value)
        for primary_key_value in primary_keys
    ]
    pipeline.delete(*names)
    # remove the primary keys from the indexz
    table_index_key = get_model_index_key(model=model)
    pipeline.zrem(table_index_key, *names)



"""Exposes the Base `Model` common to both async and sync APIs

"""
import typing
from typing import Dict, Tuple, Any, Type, Union, List, Optional

from pydantic import ConfigDict, BaseModel
from pydantic.fields import ModelPrivateAttr

from pydantic_redis._shared.utils import (
    typing_get_origin,
    typing_get_args,
    from_any_to_valid_redis_type,
    from_dict_to_key_value_list,
    from_bytes_to_str,
    from_str_or_bytes_to_any,
)


from ..store import AbstractStore


class AbstractModel(BaseModel):
    """A base class for all Models, sync and async alike.

    See the child classes for more.

    Attributes:
        _primary_key_field (str): the field that can uniquely identify each record
            for the current Model
        _field_types (Dict[str, Any]): a mapping of the fields and their types for
            the current model
        _store (AbstractStore): the Store in which the current model is registered.
        _nested_model_tuple_fields (Dict[str, Tuple[Any, ...]]): a mapping of
            fields and their types for fields that have tuples of nested models
        _nested_model_list_fields (Dict[str, Type["AbstractModel"]]): a mapping of
            fields and their associated nested models for fields that have
            lists of nested models
        _nested_model_fields (Dict[str, Type["AbstractModel"]]): a mapping of
            fields and their associated nested models for fields that have nested models
    """

    _primary_key_field: str
    _field_types: Dict[str, Any] = {}
    _store: AbstractStore
    _nested_model_tuple_fields: Dict[str, Tuple[Any, ...]] = {}
    _nested_model_list_fields: Dict[str, Type["AbstractModel"]] = {}
    _nested_model_fields: Dict[str, Type["AbstractModel"]] = {}
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def get_store(cls) -> AbstractStore:
        """Gets the Store in which the current model is registered.

        Returns:
             the instance of the store for this model
        """
        return cls._store

    @classmethod
    def get_nested_model_tuple_fields(cls):
        """Gets the mapping for fields that have tuples of nested models.

        Returns:
             The mapping of field name and field type of a form similar to
              `Tuple[str, Book, date]`
        """
        return cls._nested_model_tuple_fields

    @classmethod
    def get_nested_model_list_fields(cls):
        """Gets the mapping for fields that have lists of nested models.

        Returns:
             The mapping of field name and model class nested in that field.
        """
        return cls._nested_model_list_fields

    @classmethod
    def get_nested_model_fields(cls):
        """Gets the mapping for fields that have nested models.

        Returns:
             The mapping of field name and model class nested in that field.
        """
        return cls._nested_model_fields

    @classmethod
    def get_primary_key_field(cls):
        """Gets the field that can uniquely identify each record of current Model

        Returns:
            the field that can be used to uniquely identify each record of current Model
        """
        try:
            return cls._primary_key_field.get_default()
        except AttributeError:
            return cls._primary_key_field

    @classmethod
    def get_field_types(cls) -> Dict[str, Any]:
        """Gets the mapping of field and field_type for current Model.

        Returns:
            the mapping of field and field_type for current Model
        """
        return cls._field_types

    @classmethod
    def initialize(cls):
        """Initializes class-wide variables for performance's reasons.

        This is a performance hack that initializes variables that are common
        to all instances of the current Model e.g. the field types.
        """
        cls._field_types = typing.get_type_hints(cls)

        cls._nested_model_list_fields = {}
        cls._nested_model_tuple_fields = {}
        cls._nested_model_fields = {}

        for field, field_type in cls._field_types.items():
            try:
                # In case the annotation is Optional, an alias of Union[X, None], extract the X
                is_generic = hasattr(field_type, "__origin__")
                if (
                    is_generic
                    and typing_get_origin(field_type) == Union
                    and typing_get_args(field_type)[-1] == None.__class__
                ):
                    field_type = typing_get_args(field_type)[0]
                    is_generic = hasattr(field_type, "__origin__")

                if (
                    is_generic
                    and typing_get_origin(field_type) in (List, list)
                    and issubclass(typing_get_args(field_type)[0], AbstractModel)
                ):
                    cls._nested_model_list_fields[field] = typing_get_args(field_type)[
                        0
                    ]

                elif (
                    is_generic
                    and typing_get_origin(field_type) in (Tuple, tuple)
                    and any(
                        [
                            issubclass(v, AbstractModel)
                            for v in typing_get_args(field_type)
                        ]
                    )
                ):
                    cls._nested_model_tuple_fields[field] = typing_get_args(field_type)

                elif issubclass(field_type, AbstractModel):
                    cls._nested_model_fields[field] = field_type

            except (TypeError, AttributeError):
                pass

    @classmethod
    def serialize_partially(cls, data: Optional[Dict[str, Any]]) -> Dict[str, Any]:
        """Casts complex data types within a given dictionary to valid redis types.

        Args:
            data: the dictionary containing data with complex data types

        Returns:
            the transformed dictionary
        """
        return {key: from_any_to_valid_redis_type(value) for key, value in data.items()}

    @classmethod
    def deserialize_partially(
        cls, data: Union[List[Any], Dict[Any, Any]] = ()
    ) -> Dict[str, Any]:
        """Casts str or bytes in a dict or flattened key-value list to expected data types.

        Converts str or bytes to their expected data types

        Args:
            data: flattened list of key-values or dictionary of data to cast.
                Keeping it as potentially a dictionary ensures backward compatibility.

        Returns:
            the dictionary of properly parsed key-values.
        """
        if isinstance(data, dict):
            # for backward compatibility
            data = from_dict_to_key_value_list(data)

        parsed_dict = {}

        nested_model_list_fields = cls.get_nested_model_list_fields()
        nested_model_tuple_fields = cls.get_nested_model_tuple_fields()
        nested_model_fields = cls.get_nested_model_fields()

        for i in range(0, len(data), 2):
            key = from_bytes_to_str(data[i])
            field_type = cls._field_types.get(key)
            value = from_str_or_bytes_to_any(value=data[i + 1], field_type=field_type)

            if key in nested_model_list_fields and value is not None:
                value = _cast_lists(value, nested_model_list_fields[key])

            elif key in nested_model_tuple_fields and value is not None:
                value = _cast_tuples(value, nested_model_tuple_fields[key])

            elif key in nested_model_fields and value is not None:
                value = _cast_to_model(value=value, model=nested_model_fields[key])

            parsed_dict[key] = value

        return parsed_dict


def _cast_lists(value: List[Any], _type: Type[AbstractModel]) -> List[AbstractModel]:
    """Casts a list of flattened key-value lists into a list of _type.

    Args:
        _type: the type to cast the records to.
        value: the value to convert

    Returns:
        a list of records of the given _type
    """
    return [_type(**_type.deserialize_partially(item)) for item in value]


def _cast_tuples(value: List[Any], _type: Tuple[Any, ...]) -> Tuple[Any, ...]:
    """Casts a list of flattened key-value lists into a list of tuple of _type,.

    Args:
        _type: the tuple signature type to cast the records to
            e.g. Tuple[str, Book, int]
        value: the value to convert

    Returns:
        a list of records of tuple signature specified by `_type`
    """
    items = []
    for field_type, value in zip(_type, value):
        if issubclass(field_type, AbstractModel) and value is not None:
            value = field_type(**field_type.deserialize_partially(value))
        items.append(value)

    return tuple(items)


def _cast_to_model(value: List[Any], model: Type[AbstractModel]) -> AbstractModel:
    """Converts a list of flattened key-value lists into a list of models,.

    Args:
        model: the model class to cast to
        value: the value to cast

    Returns:
        a list of model instances of type `model`
    """
    return model(**model.deserialize_partially(value))



"""Exposes shared utilities and base classes for pydantic-redis

This includes basic functionality of mutating and querying
redis via the pydantic-redis ORM regardless of whether this
is done asynchronously or synchronously.
This is a private package.
"""



"""Exposes base Store class common to both sync and async

"""
from typing import Optional, Union, Type, Dict, Any

from pydantic.fields import ModelPrivateAttr
from redis import Redis
from redis.asyncio import Redis as AioRedis
from pydantic import ConfigDict, BaseModel
from redis.commands.core import Script, AsyncScript

from ..config import RedisConfig
from .lua_scripts import (
    SELECT_ALL_FIELDS_FOR_ALL_IDS_SCRIPT,
    SELECT_ALL_FIELDS_FOR_SOME_IDS_SCRIPT,
    SELECT_SOME_FIELDS_FOR_ALL_IDS_SCRIPT,
    SELECT_SOME_FIELDS_FOR_SOME_IDS_SCRIPT,
    PAGINATED_SELECT_ALL_FIELDS_FOR_ALL_IDS_SCRIPT,
    PAGINATED_SELECT_SOME_FIELDS_FOR_ALL_IDS_SCRIPT,
)

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .model import AbstractModel


class AbstractStore(BaseModel):
    """An abstract class of a store.

    Check the child classes for more definition.
    """

    name: str
    redis_config: RedisConfig
    redis_store: Optional[Union[Redis, AioRedis]] = None
    life_span_in_seconds: Optional[int] = None
    select_all_fields_for_all_ids_script: Optional[Union[AsyncScript, Script]] = None
    paginated_select_all_fields_for_all_ids_script: Optional[
        Union[AsyncScript, Script]
    ] = None
    select_all_fields_for_some_ids_script: Optional[Union[AsyncScript, Script]] = None
    select_some_fields_for_all_ids_script: Optional[Union[AsyncScript, Script]] = None
    paginated_select_some_fields_for_all_ids_script: Optional[
        Union[AsyncScript, Script]
    ] = None
    select_some_fields_for_some_ids_script: Optional[Union[AsyncScript, Script]] = None
    models: Dict[str, Type["AbstractModel"]] = {}
    model_config = ConfigDict(arbitrary_types_allowed=True, from_attributes=True)

    def __init__(
        self,
        name: str,
        redis_config: RedisConfig,
        redis_store: Optional[Union[Redis, AioRedis]] = None,
        life_span_in_seconds: Optional[int] = None,
        **data: Any,
    ):
        super().__init__(
            name=name,
            redis_config=redis_config,
            redis_store=redis_store,
            life_span_in_seconds=life_span_in_seconds,
            **data,
        )

        self.redis_store = self._connect_to_redis()
        self._register_lua_scripts()

    def _connect_to_redis(self) -> Union[Redis, AioRedis]:
        """Connects the store to redis.

        Connects to the redis database basing on the `redis_config`
        attribute of this instance.

        Returns:
            A connection object to a redis database
        """
        raise NotImplementedError("implement _connect_to_redis first")

    def _register_lua_scripts(self):
        """Registers the lua scripts for this redis instance.

        In order to save on memory and bandwidth, the redis lua scripts
        need to be called using EVALSHA instead of EVAL. The latter transfers
        the scripts to the redis server on every invocation while the former
        saves the script in redis itself and invokes it using a hashed (SHA) value.
        """
        self.select_all_fields_for_all_ids_script = self.redis_store.register_script(
            SELECT_ALL_FIELDS_FOR_ALL_IDS_SCRIPT
        )
        self.paginated_select_all_fields_for_all_ids_script = (
            self.redis_store.register_script(
                PAGINATED_SELECT_ALL_FIELDS_FOR_ALL_IDS_SCRIPT
            )
        )
        self.select_all_fields_for_some_ids_script = self.redis_store.register_script(
            SELECT_ALL_FIELDS_FOR_SOME_IDS_SCRIPT
        )
        self.select_some_fields_for_all_ids_script = self.redis_store.register_script(
            SELECT_SOME_FIELDS_FOR_ALL_IDS_SCRIPT
        )
        self.paginated_select_some_fields_for_all_ids_script = (
            self.redis_store.register_script(
                PAGINATED_SELECT_SOME_FIELDS_FOR_ALL_IDS_SCRIPT
            )
        )
        self.select_some_fields_for_some_ids_script = self.redis_store.register_script(
            SELECT_SOME_FIELDS_FOR_SOME_IDS_SCRIPT
        )

    def register_model(self, model_class: Type["AbstractModel"]):
        """Registers the model to this store.

        Each store manages a number of models. In order to associate
        a model to a redis database, a Store must register it.

        Args:
            model_class: the class which represents a given schema of
                a certain type of records to be saved in redis.
        """
        if not isinstance(model_class.get_primary_key_field(), str):
            raise AttributeError(
                f"{model_class.__name__} should have a _primary_key_field"
            )

        model_class._store = self
        model_class.initialize()
        self.models[model_class.__name__.lower()] = model_class

    def model(self, name: str) -> Type["AbstractModel"]:
        """Gets a model by name. This is case insensitive.

        Args:
            name: the case-insensitive name of the model class

        Returns:
            the class corresponding to the given name
        """
        return self.models[name.lower()]



"""Exposes the utilities and the base classes for models.

This includes basic functionality of mutating and querying
redis via the pydantic-redis ORM regardless of whether this
is done asynchronously or synchronously.
"""

from .base import AbstractModel



"""A simple declarative ORM for redis based on pydantic.

Provides:

1. A subclass-able `Model` class to create Object Relational Mapping to
redis hashes
2. A redis `Store` class to mutate and query `Model`'s registered in it
3. A `RedisConfig` class to pass to the `Store` constructor to connect
to a redis instance
4. A synchronous `syncio` and an asynchronous `asyncio` interface to the
above classes
5. Parent-child relationships allowing for nesting models within models.
"""

from pydantic_redis.syncio import Store, Model, RedisConfig
import pydantic_redis.asyncio

__all__ = [Store, RedisConfig, Model, asyncio]

__version__ = "0.5.0"
